package hr.pmf.doktorat.heuristic.population;

import java.io.File;
import java.io.IOException;
import java.nio.file.Files;
import java.nio.file.Paths;
import java.text.DecimalFormat;
import java.text.DecimalFormatSymbols;
import java.util.*;
import java.util.stream.Collectors;
import java.util.stream.Stream;

public class HeuristicPopulationAbsolute {

    public static void main(String[] args) throws IOException {


        File[] files = Paths.get("resources", "input").toFile().listFiles();

        for (File file : files) {
            List<String> lines = Files.readAllLines(file.toPath()).stream().filter(line -> line.trim().length() > 0).collect(Collectors.toList());
            Random random = new Random();

            int verticesNum = -1;
            int edgesNum = -1;
            int scenariosNum = -1;

            String infoLine = lines.get(0);
            String[] parts = infoLine.split("\\s+");
            verticesNum = Integer.parseInt(parts[0]);
            edgesNum = Integer.parseInt(parts[1]);
            scenariosNum = Integer.parseInt(parts[2]);

            List<Double> iterationTimes = new ArrayList<>();
            List<Long> iterationWeights = new ArrayList<>();

            for (int iteration = 0; iteration < 10; iteration++) {


                Population.WEIGHTS_NUM = scenariosNum;

                int TOTAL_POPULATIONS = 10;

                List<Node> tree = new ArrayList<>();
                for (int verticesIndex = 0; verticesIndex < verticesNum; verticesIndex++) {
                    Node node = new Node();
                    node.index = verticesIndex;
                    tree.add(node);
                }

                tree.get(0).level = 0;

                lines.subList(1, edgesNum + 1).forEach(edgeLine -> {
                    String[] edgeParts = edgeLine.split("\\s+");
                    int parentIndex = Integer.parseInt(edgeParts[1]);
                    int childIndex = Integer.parseInt(edgeParts[2]);

                    Node child = tree.get(childIndex);
                    Node parent = tree.get(parentIndex);
                    parent.children.add(child);
                    child.bitIndex = (int) Math.pow(2, parent.children.size() - 1);
                    child.level = parent.level + 1;
                    child.parent = parent;
                });

                lines.subList(edgesNum + 1, edgesNum + scenariosNum + 1).forEach(scenarioLine -> {
                    String[] scenarioParts = scenarioLine.split("\\s+");
                    for (int i = 0, scenarioPartsLength = scenarioParts.length; i < scenarioPartsLength; i++) {
                        tree.get(i).weights.add(Integer.parseInt(scenarioParts[i]));
                    }
                });

                for (Node node : tree) {
                    if (node.children.size() == 0) {
                        Population population = new Population();
                        population.members = Collections.singletonList(node);
                        setWeights(scenariosNum, population.weights, population.members);
                        population.recalculateWeights();
                        node.setIncluded.add(population);
                    }
                }

                long start = System.currentTimeMillis();
                for (int verticesIndex = (verticesNum - 1); verticesIndex >= 0; verticesIndex--) {
                    Node currentNode = tree.get(verticesIndex);

                    if (!currentNode.children.isEmpty()) {
                        //take best worst weight
                        Population bestIncludedPop = new Population();
                        for (int childIndex = 0; childIndex < currentNode.children.size(); childIndex++) {
                            Node child = currentNode.children.get(childIndex);
                            if (!child.setExcluded.isEmpty()) {
                                addPopulationToPopulation(bestIncludedPop, findBestPopulation(child.setExcluded), false);

                            }
                        }
                        if (bestIncludedPop.members.size() != 0) {
                            bestIncludedPop.addMember(currentNode);
                            currentNode.setIncluded.add(bestIncludedPop);
                        }

                        //take best average
                        Population bestAverageIncludedPop = new Population();
                        for (int childIndex = 0; childIndex < currentNode.children.size(); childIndex++) {
                            Node child = currentNode.children.get(childIndex);
                            if (!child.setExcluded.isEmpty()) {
                                addPopulationToPopulation(bestAverageIncludedPop, findBestAveragePopulation(child.setExcluded), false);

                            }
                        }
                        if (bestAverageIncludedPop.members.size() != 0) {
                            bestAverageIncludedPop.addMember(currentNode);
                            currentNode.setIncluded.add(bestAverageIncludedPop);
                        }

                        for (int popNumber = 0; popNumber < TOTAL_POPULATIONS; popNumber++) {
                            Population includedPopulation = new Population();
                            for (int i = 0; i < currentNode.children.size(); i++) {
                                Node child = currentNode.children.get(i);
                                if (!child.setExcluded.isEmpty()) {
                                    addPopulationToPopulation(includedPopulation, pickWeightedRandomPopulation(child.setExcluded, random), false);
                                }
                            }
                            includedPopulation.addMember(currentNode);
                            currentNode.setIncluded.add(includedPopulation);
                        }

                        //take best worst weight
                        Population bestExcludedPopulation = new Population();
                        for (int childIndex = 0; childIndex < currentNode.children.size(); childIndex++) {
                            Node child = currentNode.children.get(childIndex);
                            if (!child.setExcluded.isEmpty() || !child.setIncluded.isEmpty()) {
                                addPopulationToPopulation(bestExcludedPopulation, findBestPopulation(Stream.concat(child.setIncluded.stream(), child.setExcluded.stream())
                                        .collect(Collectors.toList())), false);

                            }
                        }

                        if (bestExcludedPopulation.members.size() != 0) {
                            bestExcludedPopulation.recalculateWeights();
                            currentNode.setExcluded.add(bestExcludedPopulation);
                        }

                        //take best average
                        Population bestAverageExcludedPopulation = new Population();
                        for (int childIndex = 0; childIndex < currentNode.children.size(); childIndex++) {
                            Node child = currentNode.children.get(childIndex);
                            if (!child.setExcluded.isEmpty() || !child.setIncluded.isEmpty()) {
                                addPopulationToPopulation(bestAverageExcludedPopulation, findBestAveragePopulation(Stream.concat(child.setIncluded.stream(), child.setExcluded.stream())
                                        .collect(Collectors.toList())), false);

                            }
                        }

                        if (bestAverageExcludedPopulation.members.size() != 0) {
                            bestAverageExcludedPopulation.recalculateWeights();
                            currentNode.setExcluded.add(bestAverageExcludedPopulation);
                        }

                        for (int popNumber = 0; popNumber < TOTAL_POPULATIONS; popNumber++) {
                            Population excludedPopulation = new Population();
                            for (int i = 0; i < currentNode.children.size(); i++) {
                                Node child = currentNode.children.get(i);
                                if (!child.setExcluded.isEmpty() || !child.setIncluded.isEmpty()) {
                                    addPopulationToPopulation(excludedPopulation, pickWeightedRandomPopulation(
                                            Stream.concat(child.setIncluded.stream(), child.setExcluded.stream())
                                                    .collect(Collectors.toList()),
                                            random), false);
                                }
                            }
                            if (excludedPopulation.members.size() != 0) {
                                excludedPopulation.recalculateWeights();
                                currentNode.setExcluded.add(excludedPopulation);
                            }

                        }
                    }

                    currentNode.children.forEach(child -> {
                        child.setExcluded = null;
                        child.setIncluded = null;
                    });
                }

                long end = System.currentTimeMillis();
                System.out.println("Time: " + (end - start) + "ms");

                Node root = tree.get(0);

                List<Population> allPopulations = Stream.concat(root.setExcluded.stream(), root.setIncluded.stream()).collect(Collectors.toList());


                long bestWeight = -1;
                Population bestPopulation = null;

                for (Population population : allPopulations) {
                    if (population.worstWeight > bestWeight) {
                        bestWeight = population.worstWeight;
                        bestPopulation = population;
                    }
                }

                System.out.println(file.getName() + " Best weight: " + bestWeight);

                iterationTimes.add((end - start) / 1000.0);
                iterationWeights.add(bestWeight);

            }


            DecimalFormat df = new DecimalFormat("0", DecimalFormatSymbols.getInstance(Locale.ENGLISH));
            df.setMaximumFractionDigits(340);

            Double averageTime = iterationTimes.stream().reduce(0.0, Double::sum) / 10.0;
            Double averageWeight = iterationWeights.stream().reduce(0L, Long::sum) / 10.0;

            String output = "time: " + df.format(averageTime) + "\nweight: " + df.format(averageWeight);
            output += "\ntimes: " + iterationTimes.stream().map(time -> df.format(time) + "").collect(Collectors.joining(","));
            output += "\nweights: " + iterationWeights.stream().map(weight -> df.format(weight) + "").collect(Collectors.joining(","));

            String OUTPUT_FOLDER = "output_heuristic_population_absolute";
            if (!Files.exists(Paths.get("resources", OUTPUT_FOLDER))) {
                Files.createDirectory(Paths.get("resources", OUTPUT_FOLDER));
            }
            Files.write(Paths.get("resources", "output_heuristic_2", "h_population_absolute_" + file.getName().split(".txt")[0] + "_output.txt"), output.getBytes());
        }


    }

    static Population findBestPopulation(List<Population> populations) {
        Population bestPopulation = populations.get(0);
        int bestWeight = bestPopulation.worstWeight;
        for (int populationIndex = 1; populationIndex < populations.size(); populationIndex++) {
            Population population = populations.get(populationIndex);
            if (population.worstWeight > bestWeight) {
                bestWeight = population.worstWeight;
                bestPopulation = population;
            }
        }
        return bestPopulation;
    }

    static Population findBestAveragePopulation(List<Population> populations) {
        Population bestPopulation = populations.get(0);
        double bestAverage = bestPopulation.averageWeight;
        for (int populationIndex = 1; populationIndex < populations.size(); populationIndex++) {
            Population population = populations.get(populationIndex);
            if (population.averageWeight > bestAverage) {
                bestAverage = population.averageWeight;
                bestPopulation = population;
            }
        }
        return bestPopulation;
    }

    static Population pickWeightedRandomPopulation(List<Population> population, Random random) {
        int initialSum = 0;
        int bestWeight = 0;
        int worstWeight = Integer.MAX_VALUE;
        for (int i = 0; i < population.size(); i++) {
            if (bestWeight < population.get(i).worstWeight) {
                bestWeight = population.get(i).worstWeight;
            }
            if (worstWeight > population.get(i).worstWeight) {
                worstWeight = population.get(i).worstWeight;
            }
        }

        worstWeight = worstWeight - 1;

        for (int i = 0; i < population.size(); i++) {
            initialSum += (population.get(i).worstWeight - worstWeight);
        }

        double result = random.nextDouble() * initialSum;

        int sum = 0;
        for (int i = 0; i < population.size(); i++) {
            sum += (population.get(i).worstWeight - worstWeight);
            if (result <= sum) {
                return population.get(i);
            }
        }


        throw new RuntimeException("Should never be thrown");
    }

    static void addPopulationToPopulation(Population addTo, Population addFrom, boolean recalculateWorstWeight) {
        addTo.members.addAll(addFrom.members);
        for (int weightIndex = 0; weightIndex < addTo.weights.length; weightIndex++) {
            addTo.weights[weightIndex] += addFrom.weights[weightIndex];
        }
        if (recalculateWorstWeight) {
            addTo.recalculateWeights();
        }
    }


    static void setWeights(int scenariosNum, int[] weights, List<Node> members) {
        int membersLength = members.size();
        for (int memberIndex = 0; memberIndex < membersLength; memberIndex++) {
            Node node = members.get(memberIndex);
            for (int scenarioIndex = 0; scenarioIndex < scenariosNum; scenarioIndex++) {
                weights[scenarioIndex] += node.weights.get(scenarioIndex);
            }
        }

    }

}