package hr.pmf.doktorat.heuristic.population;

import java.io.File;
import java.io.IOException;
import java.nio.file.Files;
import java.nio.file.Paths;
import java.text.DecimalFormat;
import java.text.DecimalFormatSymbols;
import java.util.*;
import java.util.stream.Collectors;
import java.util.stream.Stream;

public class HeuristicPopulationRelative {

    public static void main(String[] args) throws IOException {

        File[] files = Paths.get("resources", "input").toFile().listFiles();
        for (File file : files) {

            List<String> lines = Files.readAllLines(file.toPath()).stream().filter(line -> line.trim().length() > 0).collect(Collectors.toList());
            List<Integer> solutions =
                    Files.readAllLines(Paths.get("resources", "standard_solutions", "standard_solution_output_" + file.getName().split(".txt")[0] + ".txt"))
                            .stream()
                            .filter(line -> line.trim().length() > 0)
                            .map(line -> Integer.parseInt(line.split(",")[2]))
                            .collect(Collectors.toList());


            Random random = new Random();

            int verticesNum = -1;
            int edgesNum = -1;
            int scenariosNum = -1;

            String infoLine = lines.get(0);
            String[] parts = infoLine.split("\\s+");
            verticesNum = Integer.parseInt(parts[0]);
            edgesNum = Integer.parseInt(parts[1]);
            scenariosNum = Integer.parseInt(parts[2]);


            List<Double> iterationTimes = new ArrayList<>();
            List<Double> iterationWeights = new ArrayList<>();

            for (int iteration = 0; iteration < 10; iteration++) {


                Population.WEIGHTS_NUM = scenariosNum;

                int TOTAL_POPULATIONS = 10;

                List<Node> tree = new ArrayList<>();
                for (int verticesIndex = 0; verticesIndex < verticesNum; verticesIndex++) {
                    Node node = new Node();
                    node.index = verticesIndex;
                    tree.add(node);
                }

                tree.get(0).level = 0;

                lines.subList(1, edgesNum + 1).forEach(edgeLine -> {
                    String[] edgeParts = edgeLine.split("\\s+");
                    int parentIndex = Integer.parseInt(edgeParts[1]);
                    int childIndex = Integer.parseInt(edgeParts[2]);

                    Node child = tree.get(childIndex);
                    Node parent = tree.get(parentIndex);
                    parent.children.add(child);
                    child.bitIndex = (int) Math.pow(2, parent.children.size() - 1);
                    child.level = parent.level + 1;
                    child.parent = parent;
                });

                lines.subList(edgesNum + 1, edgesNum + scenariosNum + 1).forEach(scenarioLine -> {
                    String[] scenarioParts = scenarioLine.split("\\s+");
                    for (int i = 0, scenarioPartsLength = scenarioParts.length; i < scenarioPartsLength; i++) {
                        tree.get(i).weights.add(Integer.parseInt(scenarioParts[i]));
                    }
                });

                for (Node node : tree) {
                    if (node.children.size() == 0) {
                        Population population = new Population();
                        population.members = Collections.singletonList(node);
                        setWeights(scenariosNum, population.weights, population.members);
                        population.recalculateWeightsWithSolutionsRelative(solutions);
                        node.setIncluded.add(population);
                    }
                }

                long start = System.currentTimeMillis();
                for (int verticesIndex = (verticesNum - 1); verticesIndex >= 0; verticesIndex--) {
                    Node currentNode = tree.get(verticesIndex);

                    if (!currentNode.children.isEmpty()) {
                        Population oldBestIncludedPop = new Population();
                        oldBestIncludedPop.best = true;

                        //take best worst weight
                        Population bestIncludedPop = new Population();
                        for (int childIndex = 0; childIndex < currentNode.children.size(); childIndex++) {
                            Node child = currentNode.children.get(childIndex);
                            if (!child.setExcluded.isEmpty()) {
                                Population bestChild = findBestPopulation(child.setExcluded);
                                addPopulationToPopulation(bestIncludedPop, bestChild, false, solutions);
                                boolean foundBest = false;
                                for (int i = 0; i < child.setExcluded.size(); i++) {
                                    if (child.setExcluded.get(i).best) {
                                        addPopulationToPopulation(oldBestIncludedPop, child.setExcluded.get(i), false, solutions);
                                        foundBest = true;
                                        break;
                                    }
                                }
                                if (!foundBest) {
                                    addPopulationToPopulation(oldBestIncludedPop, bestChild, false, solutions);
                                }

                            }
                        }

                        bestIncludedPop.addMemberWithSolutionsRelative(currentNode, solutions);
                        currentNode.setIncluded.add(bestIncludedPop);


                        oldBestIncludedPop.addMemberWithSolutionsRelative(currentNode, solutions);
                        currentNode.setIncluded.add(oldBestIncludedPop);


                        //take best average
                        Population bestAverageIncludedPop = new Population();
                        for (int childIndex = 0; childIndex < currentNode.children.size(); childIndex++) {
                            Node child = currentNode.children.get(childIndex);
                            if (!child.setExcluded.isEmpty()) {
                                addPopulationToPopulation(bestAverageIncludedPop, findBestAveragePopulation(child.setExcluded), false, solutions);

                            }
                        }

                        bestAverageIncludedPop.addMemberWithSolutionsRelative(currentNode, solutions);
                        currentNode.setIncluded.add(bestAverageIncludedPop);

                        for (int popNumber = 0; popNumber < TOTAL_POPULATIONS; popNumber++) {
                            Population includedPopulation = new Population();
                            for (int i = 0; i < currentNode.children.size(); i++) {
                                Node child = currentNode.children.get(i);
                                if (!child.setExcluded.isEmpty()) {
                                    addPopulationToPopulation(includedPopulation, pickWeightedRandomPopulation(child.setExcluded, random), false, solutions);
                                }
                            }
                            if (includedPopulation.members.size() != 0) {
                                includedPopulation.addMemberWithSolutionsRelative(currentNode, solutions);
                                currentNode.setIncluded.add(includedPopulation);
                            }
                        }


                        Population oldBestExcludedPop = new Population();
                        oldBestExcludedPop.best = true;

                        //take best worst weight
                        Population bestExcludedPopulation = new Population();
                        for (int childIndex = 0; childIndex < currentNode.children.size(); childIndex++) {
                            Node child = currentNode.children.get(childIndex);
                            if (!child.setExcluded.isEmpty() || !child.setIncluded.isEmpty()) {
                                Population bestChild = findBestPopulation(Stream.concat(child.setIncluded.stream(), child.setExcluded.stream())
                                        .collect(Collectors.toList()));
                                addPopulationToPopulation(bestExcludedPopulation, bestChild, false, solutions);
                                boolean foundBest = false;
                                for (int i = 0; i < child.setExcluded.size(); i++) {
                                    if (child.setExcluded.get(i).best) {
                                        addPopulationToPopulation(oldBestExcludedPop, child.setExcluded.get(i), false, solutions);
                                        foundBest = true;
                                        break;
                                    }
                                }

                                if (!foundBest) {
                                    for (int i = 0; i < child.setIncluded.size(); i++) {
                                        if (child.setIncluded.get(i).best) {
                                            addPopulationToPopulation(oldBestExcludedPop, child.setIncluded.get(i), false, solutions);
                                            foundBest = true;
                                            break;
                                        }
                                    }
                                }
                                if (!foundBest) {
                                    addPopulationToPopulation(oldBestExcludedPop, bestChild, false, solutions);
                                }


                            }
                        }

                        bestExcludedPopulation.recalculateWeightsWithSolutionsRelative(solutions);
                        currentNode.setExcluded.add(bestExcludedPopulation);

                        oldBestExcludedPop.recalculateWeightsWithSolutionsRelative(solutions);
                        currentNode.setExcluded.add(oldBestExcludedPop);


                        //take best average
                        Population bestAverageExcludedPopulation = new Population();
                        for (int childIndex = 0; childIndex < currentNode.children.size(); childIndex++) {
                            Node child = currentNode.children.get(childIndex);
                            if (!child.setExcluded.isEmpty() || !child.setIncluded.isEmpty()) {
                                addPopulationToPopulation(bestAverageExcludedPopulation, findBestAveragePopulation(Stream.concat(child.setIncluded.stream(), child.setExcluded.stream())
                                        .collect(Collectors.toList())), false, solutions);

                            }
                        }

                        if (bestAverageExcludedPopulation.members.size() != 0) {
                            bestAverageExcludedPopulation.recalculateWeightsWithSolutionsRelative(solutions);
                            currentNode.setExcluded.add(bestAverageExcludedPopulation);
                        }

                        for (int popNumber = 0; popNumber < TOTAL_POPULATIONS; popNumber++) {
                            Population excludedPopulation = new Population();
                            for (int i = 0; i < currentNode.children.size(); i++) {
                                Node child = currentNode.children.get(i);
                                if (!child.setExcluded.isEmpty() || !child.setIncluded.isEmpty()) {
                                    addPopulationToPopulation(excludedPopulation, pickWeightedRandomPopulation(
                                            Stream.concat(child.setIncluded.stream(), child.setExcluded.stream())
                                                    .collect(Collectors.toList()),
                                            random), false, solutions);
                                }
                            }
                            if (excludedPopulation.members.size() != 0) {
                                excludedPopulation.recalculateWeightsWithSolutionsRelative(solutions);
                                currentNode.setExcluded.add(excludedPopulation);
                            }

                        }
                    }

                    currentNode.children.forEach(child -> {
                        child.setExcluded = null;
                        child.setIncluded = null;
                    });
                }

                long end = System.currentTimeMillis();
                System.out.println("Time: " + (end - start) + "ms");

                Node root = tree.get(0);

                List<Population> allPopulations = Stream.concat(root.setExcluded.stream(), root.setIncluded.stream()).collect(Collectors.toList());


                double bestWeight = Long.MAX_VALUE;
                Population bestPopulation = null;

                for (Population population : allPopulations) {
                    if (population.worstWeightRelative < bestWeight) {
                        bestWeight = population.worstWeightRelative;
                        bestPopulation = population;
                    }
                }
                DecimalFormat df = new DecimalFormat("0", DecimalFormatSymbols.getInstance(Locale.ENGLISH));
                df.setMaximumFractionDigits(340);
                System.out.println(file.getName() + " Best weight: " + df.format(bestWeight));

                iterationTimes.add((end - start) / 1000.0);
                iterationWeights.add(bestWeight);

            }

            DecimalFormat df = new DecimalFormat("0", DecimalFormatSymbols.getInstance(Locale.ENGLISH));
            df.setMaximumFractionDigits(340);

            Double averageTime = iterationTimes.stream().reduce(0.0, Double::sum) / 10.0;
            Double averageWeight = iterationWeights.stream().reduce(0.0, Double::sum) / 10.0;

            String output = "time: " + df.format(averageTime) + "\nweight: " + df.format(averageWeight);
            output += "\ntimes: " + iterationTimes.stream().map(time -> df.format(time) + "").collect(Collectors.joining(","));
            output += "\nweights: " + iterationWeights.stream().map(weight -> df.format(weight) + "").collect(Collectors.joining(","));

            String OUTPUT_FOLDER = "output_heuristic_population_relative";
            if (!Files.exists(Paths.get("resources", OUTPUT_FOLDER))) {
                Files.createDirectory(Paths.get("resources", OUTPUT_FOLDER));
            }
            Files.write(Paths.get("resources", OUTPUT_FOLDER, "h_population_relative_" + file.getName().split(".txt")[0] + "_output.txt"), output.getBytes());
            
        }
    }

    static Population findBestPopulation(List<Population> populations) {
        Population bestPopulation = populations.get(0);
        double bestWeight = bestPopulation.worstWeightRelative;
        for (int populationIndex = 1; populationIndex < populations.size(); populationIndex++) {
            Population population = populations.get(populationIndex);
            if (population.worstWeightRelative < bestWeight) {
                bestWeight = population.worstWeightRelative;
                bestPopulation = population;
            }
        }
        return bestPopulation;
    }

    static Population findBestAveragePopulation(List<Population> populations) {
        Population bestPopulation = populations.get(0);
        double bestAverage = bestPopulation.averageWeightRelative;
        for (int populationIndex = 1; populationIndex < populations.size(); populationIndex++) {
            Population population = populations.get(populationIndex);
            if (population.averageWeightRelative < bestAverage) {
                bestAverage = population.averageWeightRelative;
                bestPopulation = population;
            }
        }
        return bestPopulation;
    }

    static Population pickWeightedRandomPopulation(List<Population> population, Random random) {
        long initialSum = 0;
        double worstWeight = -1;
        for (int i = 0; i < population.size(); i++) {

            if (worstWeight < population.get(i).worstWeightRelative) {
                worstWeight = population.get(i).worstWeightRelative;
            }
        }

        worstWeight = worstWeight + 1;

        for (int i = 0; i < population.size(); i++) {
            initialSum += (worstWeight - population.get(i).worstWeightRelative);
        }

        double result = random.nextDouble() * initialSum;

        int sum = 0;
        for (int i = 0; i < population.size(); i++) {
            sum += worstWeight - population.get(i).worstWeightRelative;
            if (result <= sum) {
                return population.get(i);
            }
        }


        throw new RuntimeException("Should never be thrown");
    }

    static void addPopulationToPopulation(Population addTo, Population addFrom, boolean recalculateWorstWeight, List<Integer> solutions) {
        addTo.members.addAll(addFrom.members);
        for (int weightIndex = 0; weightIndex < addTo.weights.length; weightIndex++) {
            addTo.weights[weightIndex] += addFrom.weights[weightIndex];
        }
        if (recalculateWorstWeight) {
            addTo.recalculateWeightsWithSolutionsRelative(solutions);
        }
    }

    static void setWeights(int scenariosNum, int[] weights, List<Node> members) {
        int membersLength = members.size();
        for (int memberIndex = 0; memberIndex < membersLength; memberIndex++) {
            Node node = members.get(memberIndex);
            for (int scenarioIndex = 0; scenarioIndex < scenariosNum; scenarioIndex++) {
                weights[scenarioIndex] += node.weights.get(scenarioIndex);
            }
        }

    }

}