package hr.pmf.doktorat.heuristic.population;

import java.util.ArrayList;
import java.util.List;
import java.util.stream.Collectors;

public class Population {

    public static int WEIGHTS_NUM;

    public List<Node> members = new ArrayList<>();

    public int worstWeight;

    public double worstWeightRelative;

    public int[] weights = new int[WEIGHTS_NUM];

    public double averageWeight = 0;

    public double averageWeightRelative = 0;

    public boolean best;

    public void addMember(Node node) {
        members.add(node);
        for (int i = 0; i < node.weights.size(); i++) {
            weights[i] += node.weights.get(i);
        }
        recalculateWeights();
    }

    public void addMemberWithSolutions(Node node, List<Integer> solutions) {
        members.add(node);
        for (int i = 0; i < node.weights.size(); i++) {
            weights[i] += node.weights.get(i);
        }
        recalculateWeightsWithSolutions(solutions);
    }

    public void addMemberWithSolutionsRelative(Node node, List<Integer> solutions) {
        members.add(node);
        for (int i = 0; i < node.weights.size(); i++) {
            weights[i] += node.weights.get(i);
        }
        recalculateWeightsWithSolutionsRelative(solutions);
    }

    void recalculateWeights() {
        double sum = 0;
        worstWeight = Integer.MAX_VALUE;
        for (int i = 0; i < this.weights.length; i++) {
            sum += this.weights[i];
            if (this.weights[i] < worstWeight) {
                this.worstWeight = this.weights[i];
            }
        }
        averageWeight = sum / weights.length;
    }

    void recalculateWeightsWithSolutions(List<Integer> solutions) {
        double sum = 0;
        worstWeight = -1;
        for (int i = 0; i < this.weights.length; i++) {
            sum += solutions.get(i) - this.weights[i];
            if ((solutions.get(i) - this.weights[i]) > worstWeight) {
                this.worstWeight = ((solutions.get(i) - this.weights[i]));
            }
        }

        averageWeight = sum / weights.length;

    }

    void recalculateWeightsWithSolutionsRelative(List<Integer> solutions) {
        double sum = 0;
        worstWeightRelative = -1;
        for (int i = 0; i < this.weights.length; i++) {
            sum += (solutions.get(i) - this.weights[i]) /  (1.0 * solutions.get(i));
            if ((solutions.get(i) - this.weights[i]) /  (1.0 * solutions.get(i)) > worstWeightRelative) {
                this.worstWeightRelative = (solutions.get(i) - this.weights[i]) /  (1.0 * solutions.get(i));
            }
        }

        averageWeightRelative = sum / (1.0 * weights.length);

    }

    @Override
    public boolean equals(Object obj) {
        Population other = (Population) obj;
        return other.toString().equals(this.toString());
    }

    @Override
    public int hashCode() {
        return toString().hashCode();
    }

    @Override
    public String toString() {
        return this.members.stream().map(node -> node.index + "").collect(Collectors.joining(","));
    }
}
